# lstmAtten_datautils/lstmAtten_datautils/base_Conf.py
import torch
import os
import logging

# 配置日志
logging.basicConfig(level=logging.INFO, format='%(message)s')
logger = logging.getLogger(__name__)

class BaseConfig:
    def __init__(self, train_data_path, test_data_path, val_data_path=None, vocab_path=None, rel_path=None,
                 max_length=70, batch_size=2):
        """
        初始化配置类，定义数据处理和模型训练的超参数。

        :param train_data_path: 训练数据集文件路径 (必填)
        :param test_data_path: 测试数据集文件路径 (必填)
        :param val_data_path: 验证数据集文件路径 (可选)
        :param vocab_path: 词表文件路径 (可选), 若提供且文件存在则使用，否则从 train_data_path 自动生成
        :param rel_path: 关系映射文件路径 (可选), 提供 relation2id.txt
        :param max_length: 最大句子长度 (默认: 70，需为正整数)
        :param batch_size: 批次大小 (默认: 2，需为正整数)
        """
        self.train_data_path = train_data_path
        self.test_data_path = test_data_path
        self.val_data_path = val_data_path or test_data_path  # 默认使用 test_data_path
        self.vocab_data_path = vocab_path or os.path.join(os.path.dirname(train_data_path), 'vocab.txt')
        self.rel_data_path = rel_path or os.path.join(os.path.dirname(train_data_path), 'relation2id.txt')  # 明确指向 relation2id.txt
        self.word2id_dict = None  # 缓存词表

        # 验证并设置 max_length
        if not isinstance(max_length, int):
            logger.warning(f"max_length 应为整数，收到 {max_length}，使用默认值 70")
            self.max_length = 70
        elif max_length <= 0:
            logger.warning(f"max_length {max_length} 无效，使用默认值 70")
            self.max_length = 70
        else:
            self.max_length = max_length

        # 验证并设置 batch_size
        if not isinstance(batch_size, int):
            logger.warning(f"batch_size 应为整数，收到 {batch_size}，使用默认值 2")
            self.batch_size = 2
        elif batch_size <= 0:
            logger.warning(f"batch_size {batch_size} 无效，使用默认值 2")
            self.batch_size = 2
        else:
            self.batch_size = batch_size

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # 验证检查文件路径
        for path in [self.train_data_path, self.test_data_path, self.val_data_path, self.vocab_data_path,
                     self.rel_data_path]:
            if not os.path.exists(path):
                logger.warning(f"路径 {path} 不存在，请确保文件可用或提供正确路径！")

        # 设置 pos 非负处理映射范围
        self.pos_range = max_length - 1

        # 设置 pos 的 padding 值，目的为了补齐
        self.pos_padding_value = self.max_length * 2

        # 自动生成或加载词表
        self.get_word2id()  # 在初始化时确保词表可用

    def get_word2id(self):
        """
        获取词到ID的映射字典，缓存机制避免重复加载。
        返回：word2id_dict 字典
        描述：若 vocab_data_path 存在且有效，则加载词表；若不存在或无效，从 train_data_path 自动生成并缓存。
        """
        if self.word2id_dict is None:
            if os.path.exists(self.vocab_data_path):
                logger.info(f"使用现有词表文件: {self.vocab_data_path}")
            else:
                logger.info(f"词表文件 {self.vocab_data_path} 不存在，自动生成...")
                from .process import get_vocab
                if not os.path.exists(self.train_data_path):
                    logger.error(f"训练数据集 {self.train_data_path} 不存在，无法生成词表！")
                    raise FileNotFoundError(f"请提供有效的 train_data_path")
                get_vocab(self)
            try:
                with open(self.vocab_data_path, encoding='utf-8') as f:
                    self.word2id_dict = {word.strip(): idx for idx, word in enumerate(f)}
                logger.info(f"词表加载完成，大小: {len(self.word2id_dict)}")
            except FileNotFoundError:
                logger.error(f"词表文件 {self.vocab_data_path} 生成后仍未找到！")
                raise
            except Exception as e:
                logger.error(f"加载词表失败: {e}")
                raise
        return self.word2id_dict


# lstmAtten_datautils/lstmAtten_datautils/data_loader.py
# coding: utf-8
from torch.utils.data import DataLoader, Dataset
from .process import *
import torch
from .base_Conf import logger, BaseConfig
from sklearn.model_selection import train_test_split
import os

# 设置完整张量打印
torch.set_printoptions(threshold=float('inf'), edgeitems=1000)  # 禁用省略，显示完整张量

class MyDataset(Dataset):
    def __init__(self, data_path, baseconfig):
        """
        初始化数据集。
        :param data_path: 数据集文件路径
        :param baseconfig: Config 对象，提供超参数和路径
        描述：从指定 data_path 加载数据，调用 get_data 生成数据集。
        """
        logger.info(f"初始化数据集: {data_path}")
        self.data = get_data(data_path, baseconfig)  # 使用特定 data_path
        logger.info(f"数据集大小: {len(self.data[0])} 条样本")

    def __len__(self):
        return len(self.data[0])

    def __getitem__(self, i):
        return self.data[0][i], self.data[1][i], self.data[2][i], self.data[3][i], self.data[4][i]

def collate_fn(batch, baseconfig):
    """
    处理批次数据。
    :param batch: 批次数据列表
    :param baseconfig: Config 对象，提供 max_length
    :return: 张量和原始数据
    描述：将批次数据转换为张量，保持与单条样本处理一致，使用 sent_padding 和 pos_padding。
    注意：适用于有标签数据，单条有标签样本应通过此函数处理。
    """
    # logger.info(f"处理批次，批次大小: {len(batch)}")
    sents = [data[0] for data in batch]
    labels = [data[1] for data in batch]
    pos_e1 = [data[2] for data in batch]
    pos_e2 = [data[3] for data in batch]
    ents = [data[4] for data in batch]
    sents_ids = [sent_padding(sent, word2id(baseconfig), baseconfig) for sent in sents]
    pos_e1_ids = [pos_padding(pos, baseconfig) for pos in pos_e1]
    pos_e2_ids = [pos_padding(pos, baseconfig) for pos in pos_e2]
    try:
        sents_tensor = torch.tensor(sents_ids, dtype=torch.long)
        pos_e1_tensor = torch.tensor(pos_e1_ids, dtype=torch.long)
        pos_e2_tensor = torch.tensor(pos_e2_ids, dtype=torch.long)
        labels_tensor = torch.tensor(labels, dtype=torch.long)
        # logger.info(f"批次张量转换完成，形状: {sents_tensor.shape}")
    except RuntimeError as e:
        logger.error(f"张量转换失败: {e}")
        raise
    # print("批次原始数据:", sents, labels, pos_e1, pos_e2, ents)
    # print("批次处理后数据:", sents_tensor, pos_e1_tensor, pos_e2_tensor, labels_tensor)
    return sents_tensor, pos_e1_tensor, pos_e2_tensor, labels_tensor, sents, labels, ents

def get_loader(baseconfig):
    """
    功能：创建数据加载器。
    输入：
        baseconfig: Config 对象，提供所有路径和超参数
        (train_data_path, test_data_path, val_data_path, vocab_path, rel_path, max_length, batch_size)
    输出：字典 {dataset_type: loader}，仅包含实际存在的加载器
    注意：baseconfig.train_data_path 和 baseconfig.test_data_path 均为必填项。
    描述：根据 baseconfig 中的路径加载数据，自动生成词表，创建训练、验证和测试 DataLoader。
    """
    logger.info("创建数据加载器...")
    # 验证必填路径
    if not baseconfig.train_data_path or not baseconfig.test_data_path:
        logger.error("train_data_path 和 test_data_path 均为必填项！")
        raise ValueError("请提供有效的 train_data_path 和 test_data_path")
    if not os.path.exists(baseconfig.train_data_path):
        logger.error(f"训练数据集文件 {baseconfig.train_data_path} 不存在！")
        raise FileNotFoundError(f"请检查 {baseconfig.train_data_path}")
    if not os.path.exists(baseconfig.test_data_path):
        logger.error(f"测试数据集文件 {baseconfig.test_data_path} 不存在！")
        raise FileNotFoundError(f"请检查 {baseconfig.test_data_path}")

    # 处理数据
    train_dataset = MyDataset(baseconfig.train_data_path, baseconfig)
    val_dataset = MyDataset(baseconfig.val_data_path, baseconfig) if baseconfig.val_data_path and os.path.exists(baseconfig.val_data_path) else None
    test_dataset = MyDataset(baseconfig.test_data_path, baseconfig) if baseconfig.test_data_path and os.path.exists(baseconfig.test_data_path) else None

    # 创建 DataLoader
    loaders = {}
    if train_dataset:
        train_loader = DataLoader(dataset=train_dataset,
                                  shuffle=True,
                                  batch_size=baseconfig.batch_size,
                                  drop_last=True,
                                  collate_fn=lambda x: collate_fn(x, baseconfig))
        loaders['train'] = train_loader
        logger.info(f"训练数据加载器: {len(train_loader)} 批次, 样本数: {len(train_dataset)}")
    if val_dataset:
        val_loader = DataLoader(dataset=val_dataset,
                                shuffle=True,
                                batch_size=baseconfig.batch_size,
                                drop_last=True,
                                collate_fn=lambda x: collate_fn(x, baseconfig))
        loaders['val'] = val_loader
        logger.info(f"验证数据加载器: {len(val_loader)} 批次, 样本数: {len(val_dataset)}")
    if test_dataset:
        test_loader = DataLoader(dataset=test_dataset,
                                 shuffle=True,
                                 batch_size=baseconfig.batch_size,
                                 drop_last=True,
                                 collate_fn=lambda x: collate_fn(x, baseconfig))
        loaders['test'] = test_loader
        logger.info(f"测试数据加载器: {len(test_loader)} 批次, 样本数: {len(test_dataset)}")

    if not loaders:
        logger.error("未创建任何数据加载器，请检查路径和配置！")
        raise ValueError("无有效数据集可用")
    return loaders

if __name__ == '__main__':
    logger.info("测试数据加载流程...")
    pass
    # baseconfig = BaseConfig(
    #     train_data_path=r"C:\Lucky_dt\2_bj\bj_23AI_KGCode\chapter4_code\BiLSTM_Attention_RE\data\train.txt",
    #     test_data_path=r"C:\Lucky_dt\2_bj\bj_23AI_KGCode\chapter4_code\BiLSTM_Attention_RE\data\test.txt",
    #     val_data_path=r"C:\Lucky_dt\2_bj\bj_23AI_KGCode\chapter4_code\BiLSTM_Attention_RE\data\val.txt",
    #     vocab_path=r"C:\Lucky_dt\2_bj\bj_23AI_KGCode\chapter4_code\BiLSTM_Attention_RE\data\vocab.txt",
    #     rel_path=r"C:\Lucky_dt\2_bj\bj_23AI_KGCode\chapter4_code\BiLSTM_Attention_RE\data\relation2id.txt"
    # )
    # loaders = get_loader(baseconfig)
    # for dataset_type, loader in loaders.items():
    #     logger.info(f'{dataset_type}加载器批次数量: {len(loader)}')
    #     for index, (sents_tensor, pos_e1_tensor, pos_e2_tensor, labels_tensor, sents, labels, ents) in enumerate(loader):
    #         logger.info(f"{dataset_type}批次 {index}: 数据处理完成")
    #         break


# lstmAtten_datautils/lstmAtten_datautils/process.py
# -*- coding: utf-8 -*-
from .base_Conf import BaseConfig, logger
from itertools import chain
import torch

def relation2id(baseconfig):
    """
    功能：读取关系数据文件，生成关系名称到ID的映射字典。
    输入：
        baseconfig: Config 对象，提供 rel_data_path
    输出：关系到ID的字典 {relation: id}
    描述：从 baseconfig.rel_data_path (relation2id.txt) 读取关系文件，每行格式为 "relation_name id"，生成映射字典。
    """
    relation2id_dict = {}
    logger.info(f"加载关系映射从: {baseconfig.rel_data_path}")
    try:
        with open(baseconfig.rel_data_path, 'r', encoding='utf-8') as f:
            for line_num, line in enumerate(f, 1):
                lines = line.strip().split(" ")
                if len(lines) != 2:
                    logger.warning(f"第 {line_num} 行格式错误: {line.strip()}, 预期 'relation_name id', 跳过")
                    continue
                relation_name, relation_id = lines[0], lines[1]
                try:
                    relation2id_dict[relation_name] = int(relation_id)
                except ValueError:
                    logger.warning(f"第 {line_num} 行 ID 无效: {relation_id}, 跳过")
                    continue
        logger.info(f"关系映射加载完成，字典大小: {len(relation2id_dict)}")
    except FileNotFoundError:
        logger.error(f"关系文件 {baseconfig.rel_data_path} 未找到！请提供或创建该文件。")
        raise
    return relation2id_dict

def word2id(baseconfig):
    """
    功能：获取词到ID的映射字典。
    输入：
        baseconfig: Config 对象，提供 vocab_data_path
    输出：词到ID的字典 {word: id}
    描述：通过 baseconfig.get_word2id() 获取缓存的词表映射，避免重复加载。
    """
    return baseconfig.get_word2id()

def sent_padding(sent, word2id_dict, baseconfig):
    """
    功能：将句子中的字符转为ID，并截断或补齐到固定长度。
    输入：
        sent: 字符列表，表示句子
        word2id_dict: 词到ID映射字典
        baseconfig: Config 对象，提供 max_length
    输出：长度为 baseconfig.max_length 的ID列表
    描述：将 sent 中的每个字符转为对应的 ID，使用 'UNK' 处理未知词，超出 max_length 截断，不足则补齐 'PAD'。
    """
    ids = [word2id_dict.get(word, word2id_dict['UNK']) for word in sent]
    if len(ids) >= baseconfig.max_length:
        return ids[:baseconfig.max_length]
    return ids + [word2id_dict['PAD']] * (baseconfig.max_length - len(ids))

def pos(num, baseconfig):
    """
    功能：将相对位置差映射到非负范围，覆盖 [-pos_range, pos_range]。
    输入：
        num: 相对位置差
        baseconfig: Config 对象，提供 pos_range
    输出：映射后的非负位置ID
    描述：根据 baseconfig.pos_range（默认 max_length-1）将位置差映射到 [0, 2*pos_range]，
          num < -pos_range 返回 0，num > pos_range 返回 2*pos_range，其他情况返回 num + pos_range。
    """
    pos_range_num = baseconfig.pos_range
    if num < -pos_range_num:
        return 0
    elif num > pos_range_num:
        return 2 * pos_range_num
    return num + pos_range_num

def pos_padding(pos_ids, baseconfig):
    """
    功能：将位置序列转为非负形式，并截断或补齐到最大长度。
    输入：
        pos_ids: 位置序列
        baseconfig: Config 对象，提供 max_length 和 pos_padding_value
    输出：长度为 baseconfig.max_length 的非负位置ID列表
    描述：对 pos_ids 中的每个位置应用 pos 函数，超出 max_length 截断，不足则补齐 pos_padding_value。
    """
    pos_seq = [pos(pos_id, baseconfig) for pos_id in pos_ids]
    if len(pos_seq) >= baseconfig.max_length:
        return pos_seq[:baseconfig.max_length]
    return pos_seq + [baseconfig.pos_padding_value] * (baseconfig.max_length - len(pos_seq))

def get_data(data_path, baseconfig):
    """
    功能：读取指定路径的数据集，转换为模型需要的格式。
    输入：
        data_path: 数据集文件路径
        baseconfig: Config 对象，提供其他超参数和路径
    输出：(datas, labels, pos_e1, pos_e2, ents)
    描述：从 data_path 读取数据，每行格式为 "ent1 ent2 relation text"，生成句子、标签和位置编码。
    """
    logger.info(f"处理数据集: {data_path}")
    datas, ents, labels, pos_e1, pos_e2 = [], [], [], [], []
    relation_dict = relation2id(baseconfig)
    if not relation_dict:
        logger.error(f"关系映射字典为空，请检查 {baseconfig.rel_data_path} 内容！")
        raise ValueError("关系映射字典初始化失败")
    try:
        with open(data_path, 'r', encoding='utf-8') as f:
            for line in f:
                line_list = line.strip().split(' ', maxsplit=3)
                if len(line_list) != 4:
                    logger.warning(f"行格式错误: {line.strip()}, 跳过")
                    continue
                ent1, ent2, relation, sentence = line_list
                if relation not in relation_dict:
                    logger.warning(f"关系 {relation} 不在映射中，跳过")
                    continue
                ents.append([ent1, ent2])
                idx1, idx2 = sentence.index(ent1), sentence.index(ent2)
                sent, pos1, pos2 = [], [], []
                for idx, word in enumerate(sentence):
                    sent.append(word)
                    pos1.append(idx - idx1)
                    pos2.append(idx - idx2)
                datas.append(sent)
                pos_e1.append(pos1)
                pos_e2.append(pos2)
                labels.append(relation_dict[relation])
        logger.info(f"数据集处理完成，数据量: {len(datas)}")
    except FileNotFoundError:
        logger.error(f"数据集文件 {data_path} 未找到！")
        raise
    return datas, labels, pos_e1, pos_e2, ents

def build_vocabulary(baseconfig):
    """
    功能：从训练数据集中提取所有字符，构建词汇表。
    输入：
        baseconfig: Config 对象，提供 train_data_path
    输出：字符列表
    描述：读取 baseconfig.train_data_path 中的文本，提取唯一字符，用于生成词表，不涉及标签或位置编码。
    """
    logger.info(f"构建词汇表从: {baseconfig.train_data_path}")
    chars = set()
    try:
        with open(baseconfig.train_data_path, 'r', encoding='utf-8') as f:
            for line in f:
                line_list = line.strip().split(' ', maxsplit=3)
                if len(line_list) >= 4:
                    sentence = line_list[3]
                    chars.update(sentence)
        logger.info(f"词汇表构建完成，字符数量: {len(chars)}")
    except FileNotFoundError:
        logger.error(f"训练数据集文件 {baseconfig.train_data_path} 未找到！")
        raise
    return ['PAD', 'UNK'] + list(chars)

def get_vocab(baseconfig):
    """
    功能：从训练数据集中提取所有字符，生成词表文件。
    输入：
        baseconfig: Config 对象，提供 train_data_path 和 vocab_data_path
    输出：生成词表文件
    描述：调用 build_vocabulary 获取字符列表，保存到 baseconfig.vocab_data_path。
    """
    data_list = build_vocabulary(baseconfig)
    logger.info(f"生成词表，保存到: {baseconfig.vocab_data_path}")
    with open(baseconfig.vocab_data_path, 'w', encoding='utf-8') as fw:
        fw.write('\n'.join(data_list))
        fw.flush()
    logger.info(f"词表已保存到 {baseconfig.vocab_data_path}")

def process_single_sample(sample, baseconfig):
    """
    功能：处理单条无标签样本数据，转换为张量格式。
    输入：
        sample: 字典，包含 "text" (句子), "ent1" (实体1), "ent2" (实体2)，无 "label" 字段
        baseconfig: Config 对象，提供 max_length 和 vocab_data_path
    输出：(sents_tensor, pos_e1_tensor, pos_e2_tensor, labels_tensor, sents, pos_e1, pos_e2, ents)
    描述：为新数据预测设计，生成与批处理一致的张量，labels_tensor 始终为 None。
    示例输入：{"text": "温暖的家歌曲...", "ent1": "温暖的家", "ent2": "余致迪"}
    注意：若 sample 包含 "label" 字段，请使用批量处理 (get_loader + collate_fn)。
    """
    logger.info(f"处理单条样本: {sample.get('text', '无文本')}")
    text = sample.get('text')
    ent1 = sample.get('ent1')
    ent2 = sample.get('ent2')

    # 验证必要字段
    if not all([text, ent1, ent2]):
        logger.error("sample 必须包含 'text', 'ent1', 'ent2' 字段！")
        raise ValueError("缺少必要字段")
    if 'label' in sample:
        logger.error("单条样本检测到 'label' 字段，请使用批量处理 (get_loader + collate_fn)！")
        raise ValueError("单条样本不支持标签，请使用批量处理")

    # 模拟批处理格式
    sents = [list(text)]
    ents = [[ent1, ent2]]
    labels = None  # 无标签
    idx1 = text.index(ent1)
    idx2 = text.index(ent2)
    pos_e1 = [[i - idx1 for i in range(len(text))]]
    pos_e2 = [[i - idx2 for i in range(len(text))]]

    # 使用现有逻辑处理
    word2id_dict = word2id(baseconfig)
    sents_ids = [sent_padding(sent, word2id_dict, baseconfig) for sent in sents]
    pos_e1_ids = [pos_padding(pos, baseconfig) for pos in pos_e1]
    pos_e2_ids = [pos_padding(pos, baseconfig) for pos in pos_e2]

    try:
        sents_tensor = torch.tensor(sents_ids, dtype=torch.long)
        pos_e1_tensor = torch.tensor(pos_e1_ids, dtype=torch.long)
        pos_e2_tensor = torch.tensor(pos_e2_ids, dtype=torch.long)
        labels_tensor = None  # 始终为 None
        logger.info(f"单条样本张量转换完成，形状: (sents: {sents_tensor.shape}, pos_e1: {pos_e1_tensor.shape})")
    except RuntimeError as e:
        logger.error(f"单条样本张量转换失败: {e}")
        raise

    return sents_tensor, pos_e1_tensor, pos_e2_tensor, labels_tensor, sents, pos_e1, pos_e2, ents


# lstmAtten_datautils/setup.py
from setuptools import setup, find_packages

setup(
    name="lstmAtten_datautils_qinglanwu",
    version="0.0.1",
    author="QL.W",
    description="A utility package for processing data for LSTM Attention models in relation classification",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=[
        "torch>=1.7.0",
        "shutup>=0.2.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)